# add comments on version BWERF_2017_01_10_version2

#install.packages("randomForest","foreach","doMC","mixtools")
#
library(randomForest)
library(foreach)
library(doMC)
library(mixtools)
library(reshape2)

registerDoMC(cores=ncore)

#function:
#			Use backward elimination method to compute the importance of predictors.
#Parameters:
#			x: the predictor matrix (TF genes), each column contains the expression value of a gene
#			y: response varible vector (a pathway gene)
#			ratio: the ratio of TF genes being elimated in each elimation
#     repeats: As there are some randomness in the result of randomForest, we run randomForest "repeats" times.
#			verbose: if verbose = TRUE, the progress of elimation will be shown
#Output:
#			a named vector that contains the importance of the TFs
rf.be=function(x,y,ratio=1/10,repeats=30,verbose=FALSE){
  n.TF = ncol(x)
  IMP = numeric(n.TF)
  names(IMP) = colnames(x)
  while (ncol(x) > 10) {
    if (verbose) {
      print(paste(ncol(x)," TFs left for training random forest model"))
    }
    imp = rep(0,ncol(x))
    names(imp) = colnames(x)
    for (i in 1:repeats) {
      fit = randomForest(
        x = x,y = y,ntree = 1000,mtry = floor(sqrt(ncol(x))),importance = TRUE
      )
      imp = imp + importance(fit)[,1]
    }
    imp = imp / repeats
    imp = sort(imp)
    num.to.el = max(floor(ncol(x) * ratio),1)
    IMP[names(imp)] = imp
    x = x[,colnames(x) %in% names(imp)[(num.to.el + 1):length(imp)]]
  }
  return(IMP)
}
#function:
#			For each pathway gene, use rf.be function to compute importance of TFs to the pathway gene.
#         use multiple cores parallelization to accelerate.
#Parameters:
#			tf.data: each columne contains the expression values of a TF gene
#			pathway.data: each columne contains the expression values of a pathway gene
#			ratio: the ratio parameter used by rf.be function
#			verbose: if verbose=TRUE, the progress will be shown
#Output:    A matrix each row corresponds to a TF, each columne corresponds to a pathway gene,
#           the value in the matrix is the important of the TF to the pathway gene.  
im.mat=function(tf.data,pathway.data, ratio=1/10,verbose=FALSE){
  num.samples <- nrow(pathway.data)
  num.pathway <- ncol(pathway.data)
  num.tf <- ncol(tf.data)
  im.list =
    foreach(i = 1:num.pathway) %dopar% {
      if (verbose) {
        print(paste("pathway ",i,"/",num.pathway,"is processed",sep = ""))
      }
      im = rf.be(tf.data,pathway.data[,i], ratio = ratio,repeats=30, verbose = verbose)
      list(im = im,index.pathway = i)
    }
  im.res = matrix(0,num.tf,num.pathway)
  rownames(im.res) = colnames(tf.data)
  colnames(im.res) = colnames(pathway.data)
  for (i in 1:num.pathway) {
    j = im.list[[i]][[2]]
    a = im.list[[i]][[1]]
    im.res[names(a),j] = a
  }
  return(im.res)
}

#function:
#			Build one layer of gene regulatory network.
#Parameters:
#			tf.data: each columne contains the expression values of a TF gene
#			pathway.data: each columne contains the expression values of a pathway gene
#			ratio: the ratio parameter used by rf.be function
#     mode: either "assign" , "auto" or "noCut"
#			num.alayer: the number of TFs user want to keep.
#Output: A list that contains one layer of GRN   
oneLayer=function(tf.data,pathway.data,num.alayer=20,ratio=1/10,mode="assign",verbose=TRUE){
  im.matrix = im.mat(tf.data,pathway.data,ratio = ratio,verbose = verbose)
  num.tf = dim(tf.data)[2]
  num.pathway = dim(pathway.data)[2]
  im.tf = apply(im.matrix,1,sum)
  im.tf = im.tf[order(im.tf,decreasing = TRUE)]
  im.tf = im.tf[im.tf > 0]
  if (mode == "assign") {
    num.cut = num.alayer
  }else{
    mixmdl = normalmixEM(im.tf,k = 3)
    k = which.max(mixmdl$mu)
    n = 1
    while (which.max(mixmdl$posterior[n,]) == k) {
      n = n + 1
    }
    n = n - 1
    if (n > num.alayer * 1.5) {
      mixmdl = normalmixEM(im.tf[1:n],k = 3)
      k = which.min(mixmdl$mu)
      n = 1
      while (which.max(mixmdl$posterior[n,]) != k) {
        n = n + 1
      }
      n = n - 1
    }
    num.cut = n
  }
  tf.selected = names(im.tf)[1:num.cut]
  sorted.indices <- order(im.matrix, decreasing = TRUE)
  edge.list = data.frame(from = character(),to = character(),im = numeric())
  indx = 0
  num.tf.selected = 0
  set.tf.selected = character()
  while (num.tf.selected < num.cut) {
    indx = indx + 1
    i = sorted.indices[indx]
    im = im.matrix[i]
    row.col <- lin.to.square(i, num.tf)
    row <- row.col[1]
    col <- row.col[2]
    if (colnames(tf.data)[row] %in% tf.selected) {
      edge.list = rbind(edge.list,data.frame(
        from = colnames(tf.data)[row],to = colnames(pathway.data)[col],im = im
      ))
      if (!(colnames(tf.data)[row] %in% set.tf.selected)) {
        set.tf.selected = c(set.tf.selected, colnames(tf.data)[row])
        num.tf.selected = num.tf.selected + 1
      }
    }
  }
  edge.list[,1] = as.character(edge.list[,1])
  edge.list[,2] = as.character(edge.list[,2])
  return(edge.list)
}

#function:
#     for a matrix, change the linear order to square order.
#parameters:
#     i:  the linear order
#     nrow: the number of rows of the matrix.
#Output:
#     a vector contains two elements, containing the row order and the column order. 
lin.to.square <- function(i, nrow) {
  col <- ((i - 1) %/% nrow) + 1
  row <- ((i - 1) %% nrow) + 1
  return(c(row, col))
}


#function:
#     Build a multilayered gene regulatory network.
#parameters:
#			tf.data: each columne contains the expression values of a TF gene
#			pathway.data: each columne contains the expression values of a pathway gene
#			ratio: the ratio parameter used by rf.be function
#     mode: either "assign" or "auto"
#			nums.tf.eachlayer: the numbers of TFs in each layer.
#Output:
#     A list that contains ML-GRN    
multiLayer=function(tf.data,pathway.data,num.layers=3,ratio=1/10,mode="assign",nums.tf.eachlayer=c(20,10,5),verbose=FALSE){
  result = data.frame(
    layer = numeric(),from = character(),to = character(),im = numeric()
  )
  for (i in 1:num.layers) {
    if (verbose) {
      print(paste("current layer: ",i))
      print("---------------------------------")
    }
    alayer = oneLayer(
      tf.data,pathway.data,num.alayer = nums.tf.eachlayer[i],mode = mode,ratio =
        ratio, verbose = verbose
    )
    select.tf = unique(alayer[,1])
    n = nrow(alayer)
    result = rbind(result,data.frame(
      layer = rep(i,n),from = alayer[,1],to = alayer[,2],im = alayer[,3]
    ))
    pathway.data = tf.data[,colnames(tf.data) %in% select.tf]
    tf.data = tf.data[,!(colnames(tf.data) %in% select.tf)]
  }
  result
}

